/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source 
 * code directory.
 */
#include "pool.h"
#include "runtime.h"
#include "select.h"
#include "magic.h"
#include "module.h"
#include "list.h"

#undef pool_destroy

void pool_destroy(POOL p);


#ifdef POOL_MAINTAINANCE
struct _filepos {
	unsigned int line;
	const char *filename;
	time_t when;
};
#endif

struct pool {
	unsigned int magic;
        void *mem,*base;
        size_t size;

        struct pool *children;

	struct pool *parent;
	struct pool *next_sibling;

	int marked;

#ifdef POOL_MAINTAINANCE
        struct _filepos allocated,destroyed;

        struct mem_header *last_header;
#endif
};

#ifdef POOL_MAINTAINANCE
struct mem_header {
        size_t size;
        struct _filepos allocated;
	magic_t magic; //MAGIC_MEM_HEADER
}
#endif

struct pool_list {
	POOL pool;
	struct pool_list *next;
};

static void pool_clean_tmp();

struct pool tmp_pool_d;

POOL tmp_pool=&tmp_pool_d;

struct pool_list *pool_destroy_list;

struct pool_stats {
	unsigned long full_hits;
	unsigned long creations;
	unsigned long allocs;
	unsigned long tmp_allocs;
} pool_stats = { 0,0,0,0 };

void pool_post_func(void)
{
	struct pool_list *pn, *next;
	POOL p;
	if (!pool_destroy_list)
		return;

	pn = pool_destroy_list; 
	while(pn != NULL) {
		p = pn->pool;
		next = pn->next;
		ASSERT(p);
		ASSERT(is_magic(p,POOL_MAGIC));
		pool_destroy(p);
		free(pn);
		pn = next;
	}
	pool_destroy_list = NULL;
	return;
}

POOL pool_new_n(POOL parent, size_t size)
{
	struct pool *p;
	size_t s;
	void *ptr;
	ASSERT(size != 0);
#ifdef ENABLE_STATS
	pool_stats.creations++;
#endif
	s = size + sizeof(struct pool);
	/* increment to next 64k block */
	//s += 64-(s%64);

	ptr = malloc(s);
	
	p = (struct pool *) ptr;
	memset(p,0,s);
	p->magic = POOL_MAGIC;
	p->base = ptr;
	p->mem = ptr + sizeof(struct pool);
	p->size = s;
	p->parent = NULL;
	p->next_sibling = NULL;
	p->children = NULL;

	//if (parent) {
	//	p->parent = parent;
	//	p->next_sibling = parent->children;
	//	parent->children = p;
	//}	
	return p;	
}

void pool_destroy(POOL p)
{
	POOL tmp,o;
	ASSERT(p != tmp_pool /* Attempt to free tmp pool */);
	ASSERT(is_magic(p, POOL_MAGIC));

//	debug(NULL,"Destroying pool %x",p);

	if (p->parent) {
		/* remove us from parents list */
		o = NULL;
		tmp = p->parent->children;
		while(tmp != NULL && tmp != p) {
			ASSERT(is_magic(tmp, POOL_MAGIC));
			o = tmp;
			tmp=tmp->next_sibling;
		}
		ASSERT(tmp != NULL);
		if (o == NULL) {
			p->parent->children = p->next_sibling;
		} else {
			o->next_sibling = p->next_sibling;
		}
	}
	if (p->children) {
		/* dealloc children */
		/* for now, full recursive */
		tmp = p->children;
		while(tmp != NULL) {
			ASSERT(is_magic(tmp, POOL_MAGIC));
			o = tmp;
			tmp=tmp->next_sibling;
			pool_destroy(o);
		}
	}
	
	p->magic = 0;
	free(p);
}

void pool_idle_destroy(POOL p)
{
	struct pool_list *pn;
	ASSERT(p);
	ASSERT(p != tmp_pool /* Attempt to destroy tmp pool */);
	ASSERT(is_magic(p,POOL_MAGIC));

	ASSERT(p->marked == 0);
	p->marked = 1;

//	debug(NULL,"Scheduling pool %x for destruction",p);
	
	pn = (struct pool_list *) malloc(sizeof(struct pool_list));
	pn->next = pool_destroy_list;
	pool_destroy_list = pn;

	pn->pool = p;
	
}
static void pool_clean_tmp()
{
	if (!tmp_pool)
		return;

	memset(tmp_pool + sizeof(struct pool),0,
			tmp_pool->size-sizeof(struct pool));
	tmp_pool->mem = tmp_pool->base + sizeof(struct pool);
}

void tmp_check_pool()
{
#ifdef ENABLE_STATS
	pool_stats.tmp_allocs++;
#endif
	if (is_magic(tmp_pool, POOL_MAGIC))
		return;
	ASSERT(tmp_pool->magic == 0 /* Should be 0 here */);
	tmp_pool->base = malloc(POOL_TMP_SIZE);
	tmp_pool->magic = POOL_MAGIC;
	tmp_pool->mem = tmp_pool->base;
	tmp_pool->size = POOL_TMP_SIZE;

	ASSERT(module_check_load("select") == 0 /* Shouldn't be using tmp_pool yyet */);
	register_pre_func(pool_clean_tmp);
}

void *tmp_alloc(size_t size)
{
	return palloc(tmp_pool,size);
}

char *tmp_strdup(const char *str)
{
	size_t len;
	char *cp;
	ASSERT(str);
	len = strlen(str);
	cp = tmp_alloc(len+1);
	if (cp)
		strcpy(cp,str);
	return cp;
}
char *tmp_strdupn(const char *str,size_t n)
{
	register size_t len;
	register const char *c;
	char *cp;
	ASSERT(n >= 0);
	c = str;
	len = 0;
	while(*c != 0 && len < n) {
		c++;
		len++;
	}
	cp = tmp_alloc(len+1);
	if (cp)
		strcpy(cp,str);
	return cp;
}

char *pstrdupn(POOL p, const char *str,size_t n)
{
	register size_t len;
	register const char *c;
	char *cp;
	ASSERT(p);
	ASSERT(str);
	ASSERT(n >= 0);
	c = str;
	len = 0;
	while(*c != 0 && len < n) {
		c++;
		len++;
	}
	cp = palloc(p,len+1);
	if (cp)
		strcpy(cp,str);
	return cp;
}



char *pstrdup(POOL p, const char *str)
{
	size_t len;
	char *cp;
	ASSERT(p);
	ASSERT(str);
	len = strlen(str);
	cp = palloc(p,len+1);
	if (cp)
		strcpy(cp,str);
	return cp;
}
	
void *palloc(struct pool *p, size_t size)
{
	POOL pc;
	void *r;
	size_t left;

	if (p == tmp_pool) 
		tmp_check_pool();
	
#ifdef ENABLE_STATS
	pool_stats.allocs++;
#endif
	ASSERT(p != NULL /* Pool is NULL */);
	ASSERT(is_magic(p,POOL_MAGIC));
	
	left = p->size - (p->mem - p->base);

	if (size > left) {
#ifdef ENABLE_STATS
		pool_stats.full_hits++;
#endif
		/* Not enough memory in this block */
		/* Search children for memory */
		for (pc = p->children; pc != NULL; pc=pc->next_sibling) {
			ASSERT(is_magic(pc,POOL_MAGIC));
			left = pc->size - (pc->mem - (void *)pc);
			if (size <= left)
				break;
		}
		if (pc == NULL) {
			/* Not enough memory in any children blocks either */
			pc = pool_new_n(p,size + p->size);
		}
		return palloc(pc,size);
	}
	r = p->mem;
	p->mem += size;
	return r;
}

